# -*- coding: utf-8 -*-
# required libraries:
# heavylight
# pip install heavylight

# %%
import heavylight
from heavylight import Table
from protection_model_mo import TermAssurance
import numpy as np
import pandas as pd


# %%

basis = {
    "cost_inflation_pa": 0.02,
    "initial_expense": 500,
    "expense_pp": 10,
    "lapse_rate_pa": 0.1,
    "mort_table": Table.read_csv(r"tables/q_x_generic.csv"),
    "forward_rates": Table.read_csv(r"tables/forward_rates.csv"),
}

def create_data(pols, seed=42):

    rng = np.random.default_rng(seed)

    # override the single datapoint `pols`
    data = dict(
        sum_assured = rng.integers(10_000, 250_000, pols),
        age_at_entry = rng.integers(20, 50, pols),
        term_y = rng.integers(10, 30, pols),
        smoker_status = rng.choice(['S', 'N'], pols),
        shape = rng.choice(['level', 'decreasing'], pols),
        annual_premium = np.ones(pols),
        init_pols_if = np.ones(pols),
        extra_mortality = np.zeros(pols),
        sex = rng.choice(['F', 'M'], pols),
    )
    return data

## with LightModel we can run unoptimised, then run optimised.

opt_data = create_data(5) # small data set to run the optimiser
data = create_data(100_000) # full data set

proj = TermAssurance()
proj.data = opt_data
proj.basis = basis
proj.RunModel(240) # run unoptimised
proj.data = data # replace data
proj.RunOptimized() # run optimised
proj.df_agg
# %%
    



import numpy as np
from heavylight import LightModel
from heavylight.memory_optimized_model import agg
from typing import Union, Callable
import pytest

class SimpleModel(LightModel):

    def __init__(self, initial_pols_if: np.ndarray, mortality_rate = .01):
        super().__init__()
        self.initial_pols_if = initial_pols_if
        self.mortality_rate = mortality_rate

    def t(self, t):
        return t
    
    def num_pols_if(self, t):
        if t == 0:
            return self.initial_pols_if
        else:
            return self.num_pols_if(t - 1) - self.pols_death(t - 1) # causes exponential time complexity if uncached
    
    def pols_death(self, t):
        return self.num_pols_if(t) * self.mortality_rate

    def cashflow(self, t):
        return self.num_pols_if(t) * 100
    
    def v(self, t):
        """discount factor for time t --> time 0"""
        if t == 0:
            return 1
        else:
            return self.v(t - 1) / (1 + self.forward_rate(t))
    
    @agg(lambda x: x * 2)
    def forward_rate(self, t):
        return 0.04
    
    def pv_cashflow(self, t):
        """present value of the cashflow occuring at time t"""
        return self.cashflow(t) * self.v(t)
    
    def fib(self, t, mult_factor):
        if t <= 1:
            return t
        return mult_factor * self.fib(t - 1, mult_factor) + self.fib(t - 2, mult_factor)

def test_method_call_and_cache_retrievals():
    sm = SimpleModel(np.linspace(.1, 1, 10))
    sm.forward_rate(0)
    assert sm.forward_rate.cache[(0,)] == .04
    assert sm._cache_graph.cache['forward_rate'][(0,)] == .04
    assert len(sm._cache_graph.cache) == 1
    assert len(sm.forward_rate.cache) == 1
    sm.forward_rate(5)
    assert len(sm.forward_rate.cache) == 2
    assert len(sm._cache_graph.cache) == 1

def test_timestep_functions():
    sm = SimpleModel(np.linspace(.1, 1, 10))
    timestep_functions = sm._timestep_functions
    expected_functions = ['t', 'num_pols_if', 'pols_death', 'cashflow', 'v', 'pv_cashflow', 'forward_rate']
    assert set(timestep_functions) == set(expected_functions) # no duplicates

@pytest.mark.timeout(4)
def test_caching_speedups():
    sm = SimpleModel(np.linspace(.1, 1, 10))
    assert len(sm.num_pols_if.cache) == 0
    sm.RunModel(200)
    assert len(sm.num_pols_if.cache) == 201
    # fib did not get called because it is not single parameter time function
    assert len(sm.fib.cache) == 0
    sm.fib(200, 1.1)
    # but it is still cached
    assert len(sm.fib.cache) == 201

def test_no_kwargs(): # we do not support keyword arguments because multiple possible hashes for same function call
    sm = SimpleModel(np.linspace(.1, 1, 10))
    with pytest.raises(ValueError):
        sm.fib(5, mult_factor=1.1)

def test_reset_cache():
    sm = SimpleModel(np.linspace(.1, 1, 10))
    sm.RunModel(5)
    assert round(np.sum(sm.pols_death.cache[(0,)]), 10) == .055
    sm.mortality_rate = .02
    sm.RunModel(5)
    assert round(np.sum(sm.pols_death.cache[(0,)]), 10) == .11



from inspect import getmembers, signature
from typing import Callable, Dict, List, Union, Any
from types import MethodType
from heavylight.memory_optimized_cache import CacheGraph, CacheMethod, FunctionCall
import pandas as pd
import numpy as np


def default_agg_function(x: Any):
    """Default aggregation function for storing results."""
    if isinstance(x, np.ndarray) and issubclass(x.dtype.type, np.number):
        return np.sum(x)
    return x


class LightModel:
    """Inheriting from this class causes functions starting with a lowercase letter to be cached. 
    The caches can be cleared to rerun the same instance with different data. 
    This model can be memory optimized to reduce the memory requirements of the model without impacting results or performance. 
    """

    def __init__(self, agg_function: Union[Callable, None] = default_agg_function):
        """**Arguments**

        `agg_function`: Aggregation function for storing results. This function is applied to the return values of each method.
         In memory optimized models the cache is cleared but the aggregated results are stored for reporting. 
         If `agg_function` is not provided, the default aggregation function is used.
         If `agg_function` is `None`, no aggregated results will be provided unless overridden at a method level with `agg`.

        The implementation of the default aggregation function is as follows:

        ```python
        def default_agg_function(x: Any):
            if isinstance(x, np.ndarray) and issubclass(x.dtype.type, np.number):
                return np.sum(x)
            return x
        ```
        """
        self._cache_graph = CacheGraph()
        self._single_param_timestep_funcs: List[CacheMethod] = []
        self._funcs: Dict[MethodType, CacheMethod] = {}
        self._ran_to: Union[int, None] = None
        # happens after setting up attributes
        for method_name, method in getmembers(self):
            if (
                not method_name[0].islower()
                or method_name.startswith("_")
                or not isinstance(method, MethodType)
            ):
                continue
            method_agg_function = getattr(method, "agg_function", None)
            if not hasattr(method, "agg_function"): # only provide class-level agg_function if method-level is not provided
                method_agg_function = agg_function
            cached_method = self._cache_graph(method_agg_function)(method)
            self._funcs[method] = cached_method
            is_single_param_t = check_single_parameter_name(method, "t")
            if is_single_param_t:
                self._single_param_timestep_funcs.append(cached_method)
            setattr(self, method_name, cached_method)

    def RunModel(self, proj_len: int):
        """**Arguments**

        - `proj_len`: Projection length. All single parameter timestep functions will be run for each timestep in `range(proj_len + 1)`.
        """
        self.Clear()
        self._ran_to = proj_len
        self._run_model(proj_len)

    def RunOptimized(self):
        if self._ran_to is None:
            raise ValueError("Model has not been run yet.")
        self.ClearOptimize()
        self._run_model(self._ran_to)

    def _run_model(self, proj_len: int):
        """shared logic between RunModel and RunOptimized"""
        for t in range(proj_len + 1):
            for func in self._single_param_timestep_funcs:
                # We avoid recalling any functions that have already been cached, resolves issue #15 lewisfogden/heavylight
                if (
                    not FunctionCall(func.func.__name__, (t,))
                    in self._cache_graph.all_calls
                ):
                    func(t)


    def Clear(self):
        """Clears the cache. If the model was memory optimized, it is no longer memory optimized."""
        self._cache_graph.reset()

    def ClearOptimize(self):
        """Clears the cache. The model is memory optimized after this function is called."""
        self._cache_graph.optimize_and_reset()

    @property
    def _timestep_functions(self):
        """List the cached functions that have a single parameter `t`."""
        return [cache.func.__name__ for cache in self._single_param_timestep_funcs]

    @property
    def cache(self):
        """
        This is the cache of the model. It is a dictionary of dictionaries. 
        The outer dictionary is keyed by the function name and the inner dictionary is keyed by the arguments.
        """
        return self._cache_graph.cache

    @property
    def cache_agg(self):
        """
        The `cache` property with the `agg_function` applied to the results.
        """
        return self._cache_graph.cache_agg

    @property
    def df(self):
        """
        A dataframe with the `cache` values of all functions that have a single parameter `t`.
        """
        return self._ToDataFrame(is_agg=False)

    @property
    def df_agg(self):
        """
        The `df` property dataframe with the `agg_function` applied to the results.
        """
        return self._ToDataFrame(is_agg=True)

    def _ToDataFrame(self, param_name="t", is_agg=False):
        get_method_cache = _get_method_cache_factory(is_agg)
        filtered_cache = {}
        for method, cache_method in self._funcs.items():
            if not check_single_parameter_name(method, param_name):
                continue
            if is_agg and cache_method.agg_func is None:
                continue
            filtered_cache[method.__name__] = get_method_cache(cache_method)
        df = pd.DataFrame(filtered_cache)
        # if t is in the dataframe, move it to first position
        if "t" in df.columns:
            df.insert(0, "t", df.pop("t"))
        return df


def check_single_parameter_name(func: Callable, name: str):
    sig = signature(func)
    return len(sig.parameters) == 1 and name in sig.parameters


def agg(agg_function: Union[Callable, None]):
    """Register the storage function of a method.
    
    Used for storing aggregated results before cache eviction to reduce memory consumption.
    """

    def decorator(func: Callable):
        func.agg_function = agg_function # type: ignore
        return func

    return decorator


def _get_method_cache_factory(is_agg: bool):
    if is_agg:
        accessor: Callable[[CacheMethod], Dict] = lambda method: method.cache_agg
    else:
        accessor: Callable[[CacheMethod], Dict] = lambda method: method.cache
    return accessor



import numpy as np
import pandas as pd
import heavylight.heavytables as ht
import pytest

def test_integer_lookup():
    lookup = ht.IntegerLookup()
    arr = np.array([1,2,3])
    assert np.all(arr==lookup.get(arr))

def test_bound_integer_lookup():
    lookup = ht.BoundIntLookup(0, 10)
    arr = np.array([-1, 0, 1, 10, 11])
    assert np.all(np.array([0, 0, 1, 10, 10])==lookup.get(arr))

def test_rectify():
    df1 = pd.DataFrame({'a': [1, 2], 'b': [1, 2], 'c': [1, 2]})
    expected_rectified_df1 = pd.DataFrame({'a': [1, 1, 2, 2], 'b': [1, 2, 1, 2], 'c': [1, np.nan, np.nan, 2]})
    assert ht.Table.rectify(df1).equals(expected_rectified_df1)

def test_string():
    df = pd.DataFrame({'key|str': ['A', 'B', 'C'], 'val|str': ['a', 'b', 'c']})
    table = ht.Table(df)
    assert table['A'] == 'a'
    assert table['C'] == 'c'
    with pytest.raises(KeyError, match=r"'invalid string key\(s\) passed into table lookup.'"):
        table['AB']
    assert list(table[np.array(['A', 'C'])]) == ['a', 'c']
    with pytest.raises(KeyError, match=r"'invalid string key\(s\) passed into table lookup.'"):
        table[np.array(['A', 'AB'])]
